#!/bin/sh
# \
exec fltkwish "$0" -- ${1+"$@"}
#
# --- quadratic.tcl --- Demonstration of solution to quadratics
#
# Copyright(C) I.B.Findleton, 2003. All Rights Reserved
#
# This script takes the 3 parameters for a quadratic equation and
# computes the roots of the equation. The curve of the parabola is
# then plotted using an XYPlot widget. Complex arithmetic is not
# implemented in this script, so imaginary roots are not tested.
#
catch { Destroy t }

# Set some initial values

set a 1
set b -8
set c -50
set r1 ""
set r2 ""
set mx 0
set my 0
set lasttag ""

# Controls for the scaling of the plot and the range of points

set xscale 1
set yscale 1
set pts 8

# A status message

set Status ""

# Create the GUI

set f [Package t.all -orient vertical]

set f1 [Package $f.parm -orient horizontal]

Option add LabeledInput.ratio .4
Option add LabeledInput.width 140
Option add LabeledInput.command "Solve $f.plot"

Option add LabeledText.ratio .6
Option add LabeledText.width 170

Option add Counter.command "Solve $f.plot"
Option add Counter.min -100
Option add Counter.step 0.5

# This is an example of a compound widget that makes use of qualified
# option names to do initialization...

proc LabeledCounter { w args } {

	global a b c pts

	set f [Package $w -orientation horizontal]

	Label $f.label -relief flat -width 70 -align left,inside
	Counter $f.counter

	eval { $f.label set } $args
	eval { $f.counter set } $args 

	return $f
	}

LabeledCounter $f1.a -Label.label "Value of A" -Counter.variable a  
LabeledCounter $f1.b -Label.label "Value of B" -Counter.variable b 
LabeledCounter $f1.c -Label.label "Value of C" -Counter.variable c

set f2 [Package $f.roots -orient horizontal]

LabeledCounter $f2.pts -Label.label "Points to plot" -Counter.variable pts \
		-Counter.min 0 -Counter.step 1
LabeledText $f2.r1 -label "Value of r1" -variable r1
LabeledText $f2.r2 -label "Value of r2" -variable r2

set f5 [Package $f.extremum -orient horizontal]
LabeledText $f5.mx -label "Extrema X" -variable mx
LabeledText $f5.my -label "Extrema Y" -variable my

Option add Scrollbar.orientation horizontal
Option add Scrollbar.min 1
Option add Scrollbar.max 100
Option add Scrollbar.step 1
Option add Scrollbar.sliderstyle nice
Option add Scrollbar.command "Solve $f.plot"

set f4 [Package $f.control -orient horizontal]
Label $f4.label -text "Horizontal Scale Factor" -w 140 -relief flat -align left,inside
Scrollbar $f4.x -variable xscale -w 400
#Scrollbar $f.y -variable yscale

# Create a plot widget. Show the least squares regression model for the
# points plotted on the graph.

XYPlot $f.plot -fit false -align top,inside -label "Plot of Quadratic" -line true \
	-xlabel "Value of X" -ylabel "Value of Y" -grid true \
	-bg blue -fg white -font helv,italic -zerox true -zeroy true

# Create a status window

Output $f.status -variable Status

set f3 [Package $f.actions -orient horizontal]
Button $f3.solve -command "Solve $f.plot" -label Solve
Button $f3.extremum -command "PlotExtremum $f.plot" -label Extremum
Button $f3.quit -label Quit -command Exit

proc Msg { what { color gray }} {

	global Status f 

#	puts $what
	$f.status set -bg $color
	set Status $what
	}

# Calculate the value of the equation for a given value of x

proc Eqn { x } {

	global a b c

	return [expr $a * $x * $x + $b * $x + $c ]
	
	}

# Make nice display values

proc Fmt { v } {

	return [format "%6.2f" $v]
	}

# Get the real part of a complex number
	
proc Real { what } {

	regsub -all { +} [string trim $what] { } what
	set list [split $what]
	
	return [lindex $list 0]
	}
	
# Get the imaginary part of a complex number

proc Imaginary { what } {

	regsub -all { +} [string trim $what] { } what
	set list [split $what]
	
	if { [llength $list] > 1 } {
		set value [lindex $list 2]
	
		regsub -all {i} $value {} value
	
		if { [lindex $list 1] == "-" } {
			return [expr -1.0 * $value]
		} else {
			return $value
			}
	} else {
		return 0
		}
	}
	
# Make a complex number from its component values

proc MakeComplex { real imaginary } {

	if { $imaginary < 0 } {
		set imag [expr -1.0 * $imaginary]
		return "$real - ${imag}i"
	} elseif { $imaginary == 0 } {
		return $real
	} else {
		return "$real + ${imaginary}i"
		}
	}
	
# Add 2 complex numbers

proc AddComplex { a b } {

	set real [expr [Real $a] + [Real $b]]
	set imaginary [expr [Imaginary $a] + [Imaginary $b]]
	
	return [MakeComplex $real $imaginary]
	}

# Subtract 2 complex numbers

proc SubtractComplex { a b } {

	set real [expr [Real $a] - [Real $b]]
	set imaginary [expr [Imaginary $a] - [Imaginary $b]]
	
	return [MakeComplex $real $imaginary]
	}
	
# Multiply 2 complex numbars

proc MultiplyComplex { a b } {

	set real [expr [Real $a] * [Real $b] - [Imaginary $a] * [Imaginary $b]]
	set imaginary [expr 2 * [Imaginary $a] * [Imaginary $b]]
	
	return [MakeComplex $real $imaginary]
	}
	
# Solve the equation and plot the results

proc Solve { w } {

	if { [catch { DoSolve $w } result] } {
		Msg "Error : $result" orange
		}
	}
	
# Do the solution

proc DoSolve { w } {

	global a b c r1 r2 disc xscale yscale pts
	global f2 mx my

	Msg "Solving y = ${a}x**2 + ${b}x + c ..."

	$w clear
	
	set r1 "" ; set r2 ""

# Handle some error conditions...

	if { $a == 0 } {
		Msg "A may not be 0 in a quadratic equation" orange
		PlotEquation $w
		return 0
		}

# Compute the value under the square root sign (discriminant)

	set disc [expr $b * $b - 4 * $a * $c]

	Msg "Discriminant = $disc"

# Compute the roots according to the value of the discriminant

	if { $disc < 0 } {
		Msg "Quadratic has imaginary roots..."
		$f2.r1 set -textbackground orange
		$f2.r2 set -textbackground orange
		set d [expr -1 * $disc]
		set d [Fmt [expr sqrt($d) / (2 * $a)]]
		if { $d > 0 } {
			set r1 "[Fmt [expr -$b / (2 * $a)]] + ${d}i"
			set r2 "[Fmt [expr -$b / (2 * $a)]] - ${d}i"
		} else {
			set d [expr -1.0 * $d]
			set r1 "[Fmt [expr -$b / (2 * $a)]] - ${d}i"
			set r2 "[Fmt [expr -$b / (2 * $a)]] + ${d}i"
			}
	} elseif { $disc == 0 } {
		$f2.r1 set -textbackground gray
		$f2.r2 set -textbackground gray
		Msg "Quadratic has equal real roots..."
		if { $b == 0 } {
			set r1 0
		} else {
			set r1 [expr -$b / (2 * a)]
			}
		set r2 $r1
	} else {
		$f2.r1 set -textbackground gray
		$f2.r2 set -textbackground gray
		Msg "Quadratic has real roots..."
		set r1 [Fmt [expr (-$b + sqrt($disc)) / (2 * $a)]]
		set r2 [Fmt [expr (-$b - sqrt($disc)) / ( 2 * $a)]]
		}

	Msg "Roots are $r1 and $r2"

# Compute the location of the extreme point

	set mx [format %6.1f [expr -$b / (2.0 * $a)]]
	set my [format %6.1f [expr $a * $mx * $mx + $b * $mx + $c]]

# Test the value of the roots in the equation

	if { $disc < 0 } {
		set y1 [AddComplex [AddComplex [MultiplyComplex $a $r1] [MultiplyComplex $b $r1]] $c]
		set y2 [AddComplex [AddComplex [MultiplyComplex $a $r2] [MultiplyComplex $b $r2]] $c]
		Msg "Test : For x = $r1 y = $y1, for x = $r2 y = $y2"
	} else {
		set y1 [expr $a * $r1 * $r1 + $b * $r1 + $c]
		set y2 [expr $a * $r2 * $r2 + $b * $r2 + $c]
		Msg "Test : For x = $r1 y = [format %6.1f $y1], for x = $r2 y = [format %6.1f $y2]"
		}
		
	PlotEquation $w
	}

# Build up a title for the plot

proc PlotEquation { w } {

	global a b c xscale yscale pts
	
	set title "Plot of Y = "

	if { $a == 1 } { 
		append title "X**2 "
	} elseif { $a == 0 } {
	} elseif { $a == -1 } {
		append title "-X**2"
	} else {
		append title "${a}X**2 "
		} 
	if { $b < 0 } {
		append title "${b}X "
	} elseif { $b > 0 } {
		append title "+ ${b}X "
		}
	if { $c < 0 } {
		append title "$c"
	} elseif { $c > 0 } {
		append title "+ $c"
		}

	$w set -title $title

#
# Add some points to the plot. Here the function is the quadratic being
# solved...
#
	set start [expr -$xscale * $pts]
	set end [expr $xscale * $pts]
	set delta $xscale

	if { $a < 0 } {
		set align bottom
	} else {
		set align top
		}

	for { set i $start } { $i <= $end } { incr i $delta } {
		set v [Eqn $i]
		$w add -x $i -y $v -color red -label $v -symbol plus \
			-linestyle dash -tag $i -labelalign $align
		}
	}

# Plot the line through the minimum or maximum

proc PlotExtremum { w } {

	global mx my

	set xo [$w set -pagex $mx]
	set yo [$w set -pagey $my]

	set len [lindex [split [$w get -pagegeometry]] 3]

	$w set -drawing "cs pc red sp $xo $yo cr 10 sp $xo 0 rt 180 fd $len"
	}

# Highight a selected point on the curve

proc HighlightPoint { w button x y } {

	set list [lindex [$w closest $x $y] 1]

	set xo [$w set -pagex [lindex $list 0]]
	set yo [$w set -pagey [lindex $list 1]]

	$w set -drawing "cs pc red sp $xo $yo cr 10"
	}

# Handle mouse motion over the drawing

proc MouseMotion { w x y } {

	global lasttag

	set tag [lindex [$w closest $x $y] 2]

	if { $lasttag != "" && $lasttag != $tag } {
		$w labelcolor black $lasttag
	} else {
		$w labelcolor orange $tag
		}

	set lasttag $tag
	}
	
# Bind an action to the click of a mouse button
	
Bind t.all.plot <ButtonPress> "HighlightPoint %W %b %x %y"

# Bind an action to mouse motion

Bind t.all.plot <Motion> "MouseMotion %W %x %y"

# Flashy appearance...

Bind Counter <Enter> "%W set -fg orange"
Bind Counter <Leave> "%W set -fg black"
Bind Button <Enter> "%W set -fg orange"
Bind Button <Leave> "%W set -fg black"

Solve $f.plot
#
Show t
#
Wm title t "Quadratic Curve Demonstration"

